using System.Diagnostics;
using System.Data;
using System.Collections;
using Microsoft.VisualBasic;
using System.Collections.Generic;
using System;
using Microsoft.VisualBasic.CompilerServices;


namespace Optim_DFP1
{
	public class MatrixLibVb
	{
		private const double DEF_BAD_VAL = -1.1e+100;
		private static double mfBadDataVal;
		private static int mnErrCount;
		private static string[] msErrMessage;
		private static string[] msErrSrc;
		private static long[] m_lPower2 = new long[32];
        static bool RandNorm_flag;
        static double RandNorm_gset;
		
		private static void INC(ref int n, int i)
		{
			n = n + i;
		}
		
		private static void DEC(ref int n, int i)
		{
			n = n - i;
		}
		
		private static void INCL(ref long n, long i)
		{
			n = n + i;
		}
		
		private static void DECL(ref long n, long i)
		{
			n = n - i;
		}
		
		private static void INCX(ref double n, double i)
		{
			n = n + i;
		}
		
		private static void DECX(ref double n, double i)
		{
			n = n - i;
		}
		
		public static void ResetErrors()
		{
			
			mnErrCount = 0;
			msErrMessage = new string[mnErrCount + 1];
			msErrSrc = new string[mnErrCount + 1];
		}
		
		public static int GetErrorCount()
		{
            return mnErrCount;
		}
		
		
		private static void ReportError(string sMessage, string sSource)
		{
			
			// adjust arrays
			msErrMessage = (string[]) Microsoft.VisualBasic.CompilerServices.Utils.CopyArray((Array) msErrMessage, new string[mnErrCount + 1]);
			msErrSrc = (string[]) Microsoft.VisualBasic.CompilerServices.Utils.CopyArray((Array) msErrSrc, new string[mnErrCount + 1]);
			msErrMessage[mnErrCount] = sMessage;
			msErrSrc[mnErrCount] = sSource;
			INC(ref mnErrCount, 1);
		}
		
		
		public static bool GetRecentMessageAndSource(ref string sErrMsg, ref string sErrSrc, bool bDecrementErrMessageIndex)
		{
			bool returnValue;
			
			// set default Exit Sub values
			sErrMsg = "";
			sErrSrc = "";
			if (mnErrCount > 0)
			{
				sErrMsg = msErrMessage[mnErrCount - 1];
				sErrSrc = msErrSrc[mnErrCount - 1];
				if (bDecrementErrMessageIndex && mnErrCount > 1)
				{
					DEC(ref mnErrCount, 1);
				}
				returnValue = true;
			}
			else
			{
				returnValue = false;
			}
			return returnValue;
		}
		
		public static void GetErrorMessages(ref string[] sMsgArr)
		{
			int i;

            sMsgArr = new string[mnErrCount + 1];
			for (i = 0; i < mnErrCount; i++)
			{
				sMsgArr[i] = msErrMessage[i];
			}
		}
		
		public static void GetErrorSources(ref string[] sMsgSrc)
		{
			int i;

            sMsgSrc = new string[mnErrCount + 1];
			for (i = 0; i < mnErrCount; i++)
			{
				sMsgSrc[i] = msErrSrc[i];
			}
		}
		
		public static double GetBadDataVal()
		{
			double returnValue;
			returnValue = mfBadDataVal;
			return returnValue;
		}
		
		
		public static double SetBadDataVal(double X)
		{
            double Y;
			
			Y = mfBadDataVal;
			mfBadDataVal = X;
			return Y;
		}
		
		private static double Max(double X, double Y)
		{
			return (X > Y ? X : Y);
		}
		
		private static double Square(double X)
		{
			return X * X;
		}
		
		private static double Sign(double A, double B)
		{
			return (B >= 0 ? (A >= 0 ? A : - A) : (A >= 0 ? - A : A));
		}
		
		private static void Swap(ref double X, ref double Y)
		{
			double temp;
			
			temp = X;
			X = Y;
			Y = temp;
		}
		
		public static long RShift(long lThis, long lBits)
		{
			long returnValue = 0;
			if (lBits <= 0)
			{
				returnValue = lThis;
			}
			else if (lBits > 63)
			{
				// .. error ...
			}
			else if (lBits > 31)
			{
				returnValue = 0;
			}
			else
			{
				if ((lThis & m_lPower2[31 - lBits]) == m_lPower2[31 - lBits])
				{
					returnValue = (lThis & (m_lPower2[31 - lBits] - 1)) * m_lPower2[lBits] | m_lPower2[31];
				}
				else
				{
					returnValue = (lThis & (m_lPower2[31 - lBits] - 1)) * m_lPower2[lBits];
				}
			}
			return returnValue;
		}
		
		public static long LShift(long lThis, long lBits)
		{
			long returnValue = 0;
			
			if (lBits <= 0)
			{
				returnValue = lThis;
			}
			else if (lBits > 63)
			{
				// ... error ...
			}
			else if (lBits > 31)
			{
				returnValue = 0;
			}
			else
			{
				if ((lThis & m_lPower2[31]) == m_lPower2[31])
				{
					returnValue = (lThis & 0x7FFFFFFF) / m_lPower2[lBits] | m_lPower2[31 - lBits];
				}
				else
				{
					returnValue = lThis / m_lPower2[lBits];
				}
			}
			return returnValue;
		}
		
		public static void InitShiftData()
		{
			m_lPower2[0] = 0x1;
			m_lPower2[1] = 0x2;
			m_lPower2[2] = 0x4;
			m_lPower2[3] = 0x8;
			m_lPower2[4] = 0x10;
			m_lPower2[5] = 0x20;
			m_lPower2[6] = 0x40;
			m_lPower2[7] = 0x80;
			m_lPower2[8] = 0x100;
			m_lPower2[9] = 0x200;
			m_lPower2[10] = 0x400;
			m_lPower2[11] = 0x800;
			m_lPower2[12] = 0x1000;
			m_lPower2[13] = 0x2000;
			m_lPower2[14] = 0x4000;
			m_lPower2[15] = 0x8000;
			m_lPower2[16] = 0x10000;
			m_lPower2[17] = 0x20000;
			m_lPower2[18] = 0x40000;
			m_lPower2[19] = 0x80000;
			m_lPower2[20] = 0x100000;
			m_lPower2[21] = 0x200000;
			m_lPower2[22] = 0x400000;
			m_lPower2[23] = 0x800000;
			m_lPower2[24] = 0x1000000;
			m_lPower2[25] = 0x2000000;
			m_lPower2[26] = 0x4000000;
			m_lPower2[27] = 0x8000000;
			m_lPower2[28] = 0x10000000;
			m_lPower2[29] = 0x20000000;
			m_lPower2[30] = 0x40000000;
			m_lPower2[31] = 0x80000000;
		}
		
		
		public static bool isFullOfEqualTo(double[,] A, double fValue)
		{			
			int nRow, nCol;
			
			for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
			{
				for (nCol = 0; nCol <= A.GetUpperBound(1); nCol++)
				{
					if (fValue != A[nRow, nCol])
					{
						return false;
					}
				}
			}
			return true;
		}
		
		public static bool isFullOfEqOrGreaterThan(double[,] A, double fValue)
		{
			int nRow, nCol;

            for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
			{
                for (nCol = 0; nCol <= A.GetUpperBound(1); nCol++)
				{
					if (A[nRow, nCol] < fValue)
					{
						return false;
					}
				}
			}
			return true;
		}
		
		
		public static bool isFullOfGreaterThan(double[,] A, double fValue)
		{
            int nRow, nCol;

            for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
			{
                for (nCol = 0; nCol <= A.GetUpperBound(1); nCol++)
				{
					if (A[nRow, nCol] <= fValue)
					{
						return false;
					}
				}
			}
			return true;
		}
		
		
		public static bool isFullOfEqOrLessThan(double[,] A, double fValue)
		{
            int nRow, nCol;

            for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
			{
                for (nCol = 0; nCol <= A.GetUpperBound(1); nCol++)
				{
					if (A[nRow, nCol] > fValue)
					{
						return false;
					}
				}
			}
			return true;
		}
		
		
		public static bool isFullOfLess(double[,] A, double fValue)
		{
            int nRow, nCol;
			
			for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
			{
				for (nCol = 0; nCol <= A.GetUpperBound(1); nCol++)
				{
					if (A[nRow, nCol] >= fValue)
					{
						return false;
					}
				}
			}
			return true;
		}
		
		public static void MatchMatDim(ref double[,] SourceMat, ref double[,] TargetMat)
		{
			TargetMat = new double[SourceMat.GetUpperBound(0) + 1, SourceMat.GetUpperBound(1) + 1];
		}
		
		public static void MatchVectDim(ref double[] SourceVect, ref double[] TargetVect)
		{
			TargetVect = new double[SourceVect.GetUpperBound(0) + 1];
		}
		
		public static void DuplicateMat(ref double[,] SourceMat, ref double[,] TargetMat)
		{
            int i, j;

            try
			{
                TargetMat = new double[SourceMat.GetUpperBound(0) + 1, SourceMat.GetUpperBound(1) + 1];
                for (i = 0; i <= SourceMat.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= SourceMat.GetUpperBound(1); j++)
					{
						TargetMat[i, j] = SourceMat[i, j];
					}
				}
				
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "DuplicateMat");
			}
		}
		
		public static void DuplicateVector(ref double[] SourceVect, ref double[] TargetVect)
		{
			int nRows;
			int i;
			
			try
			{
				nRows = SourceVect.GetUpperBound(0)+1;
				TargetVect = new double[nRows];
				for (i = 0; i < nRows; i++)
				{
					TargetVect[i] = SourceVect[i];
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "DuplicateVect");
			}
		}
		
		public static void MatCopy(ref double[,] SourceMat, ref double[,] TargetMat)
		{
			// Copy BUT DOES NOT RESIZE target matrix
			int nRows;
			int nCols;
			int i;
			int j;
			
			try
			{
				nRows = SourceMat.GetUpperBound(0) + 1;
                nCols = SourceMat.GetUpperBound(1) + 1;
				for (i = 0; i < nRows; i++)
				{
					for (j = 0; j < nCols; j++)
					{
						TargetMat[i, j] = SourceMat[i, j];
					}
				}				
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatCopy");
			}
		}
		
		public static void MatPartialCopy(ref double[,] SourceMat, ref double[,] TargetMat, int nRowStart, int nNumRows, int nColStart, int nNumCols, int nRowStartTarg, int nColStartTarg)
		{
			
			int nRowEnd, nColEnd, iCol, iRow, i, j;
			
			try
			{
				nRowEnd = nRowStart + nNumRows - 1;
				nColEnd = nColStart + nNumCols - 1;
				
				if ((nRowStart < 0) || (nNumRows < 1) || (nRowEnd < 0) || 
                    (nColStart < 0) || (nNumCols < 1) || (nColEnd < 0))
				{
					ReportError("Invalid indices", "MatPartialCopy");
					return;
				}
				
				iRow = nRowStartTarg;
				for (i = nRowStart; i <= nRowEnd; i++)
				{
					iCol = nColStartTarg;
					for (j = nColStart; j <= nColEnd; j++)
					{
						TargetMat[iRow, iCol] = SourceMat[i, j];
						iCol++;
					}
					iRow++;
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatPartialCopy");
			}
		}
		
		public static void CopySomeVectElemnts(ref double[] SourceVect, ref double[] TargetVect, int nFirst, int nNumElems, int nFirstTarg)
		{
			
			int i, nLast;
			
			try
			{
				nLast = nFirst + nNumElems - 1;
				
				if ((nFirst < 0) ||(nNumElems < 1) || (nLast < 0))
				{
					ReportError("Invalid indices", "CopySomeVectElemnts");
					return;
				}
				
				for (i = nFirst; i <= nLast; i++)
				{
					TargetVect[nFirstTarg] = SourceVect[i];
					nFirstTarg++;
				}				
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "CopySomeVectElemnts");
			}
		}
		
		public static int VectorFindVal(double[] AVect, double SearchVal, double MinDiff, int nFirst)
		{
			int i;
			
			if (nFirst < 0)
			{
				nFirst = 0;
			}
			
			for (i = nFirst; i < AVect.Length; i++)
			{
				if (Math.Abs(AVect[i] - SearchVal) < MinDiff)
				{
					return i;
				}
			}
			return -1;
		}
		
		public static int VectorFindValRev(double[] AVect, double SearchVal, double MinDiff, int nFirst)
		{
			int i;
			
			if (nFirst > (AVect.Length - 1))
			{
				nFirst = AVect.Length - 1;
			}
			
			for (i = nFirst; i >= 0; i--)
			{
				if (Math.Abs(AVect[i] - SearchVal) < MinDiff)
				{
					return i;
				}
			}
			return -1;
		}
		
		public static bool MatFindValByRows(double[,] A, double SearchVal, int StartRow, int StartCol, ref int RowFound, ref int ColFound, double MinDiff)
		{
			// search row by row
			int Row, Col;
			
			if (StartRow < 0)
			{
				StartRow = 0;
			}
			if (StartCol < 0)
			{
				StartCol = 0;
			}
			// search StartRow first
			for (Col = StartCol; Col <= A.GetUpperBound(1); Col++)
			{
				if (Math.Abs(A[StartRow, Col] - SearchVal) < MinDiff)
				{
					RowFound = StartRow;
					ColFound = Col;
					return true;
				}
			}
			// search the entire subsequent rows
			for (Col = 0; Col <= A.GetUpperBound(1); Col++)
			{
				for (Row = StartRow + 1; Row <= A.GetUpperBound(0); Row++)
				{
					if (Math.Abs(A[Row, Col] - SearchVal) < MinDiff)
					{
						RowFound = Row;
						ColFound = Col;
						return true;
					}
				}
			}
			RowFound = - 1;
			ColFound = - 1;
			return false;
		}
		
		public static bool MatFindValByRowsRev(double[,] A, double SearchVal, int StartRow, int StartCol, ref int RowFound, ref int ColFound, double MinDiff)
		{
			// reverse search row by row. Starts with higher-index rows and work to
			// low-index rows
			int Row, Col;
			
			if (StartRow > A.GetUpperBound(0))
			{
                StartRow = A.GetUpperBound(0);
			}
            if (StartCol > A.GetUpperBound(1))
			{
				StartCol = A.GetUpperBound(1);
			}
			// search StartRow first
			for (Col = StartCol; Col >= 0; Col--)
			{
				if (Math.Abs(A[StartRow, Col] - SearchVal) < MinDiff)
				{
					RowFound = StartRow;
					ColFound = Col;
					return true;
				}
			}
			// search the entire previous rows
			for (Col = A.GetUpperBound(1); Col >= 0; Col--)
			{
				for (Row = StartRow - 1; Row >= 0; Row--)
				{
					if (Math.Abs(A[Row, Col] - SearchVal) < MinDiff)
					{
						RowFound = Row;
						ColFound = Col;
						return true;
					}
				}
			}
			RowFound = -1;
			ColFound = -1;
			return false;
		}
		
		
		public static bool MatFindValByCols(double[,] A, double SearchVal, int StartRow, int StartCol, ref int RowFound, ref int ColFound, double MinDiff)
		{
			// search column by column
			int Row, Col;
			
			if (StartRow < 0)
			{
				StartRow = 0;
			}
			if (StartCol < 0)
			{
				StartCol = 0;
			}
			// search StartRow first
			for (Row = StartRow; Row <= A.GetUpperBound(0); Row++)
			{
				if (Math.Abs(A[Row, StartCol] - SearchVal) < MinDiff)
				{
					RowFound = Row;
					ColFound = StartCol;
					return true;
				}
			}
			// search the entire subsequent columns
			for (Row = 0; Row <= A.GetUpperBound(0); Row++)
			{
				for (Col = StartCol + 1; Col <= A.GetUpperBound(1); Col++)
				{
					if (Math.Abs(A[Row, Col] - SearchVal) < MinDiff)
					{
						RowFound = Row;
						ColFound = Col;
						return true;
					}
				}
			}
			RowFound = -1;
			ColFound = -1;
			return false;
		}
		
		public static bool MatFindValByColsRev(double[,] A, double SearchVal, int StartRow, int StartCol, ref int RowFound, ref int ColFound, double MinDiff)
		{
			// backward search column by column
			int Row, Col;
			
			if (StartRow < 0)
			{
				StartRow = 0;
			}
			if (StartCol < 0)
			{
				StartCol = 0;
			}
			// search StartRow first
			for (Row = StartRow; Row >= 0; Row--)
			{
				if (Math.Abs(A[Row, StartCol] - SearchVal) < MinDiff)
				{
					RowFound = Row;
					ColFound = StartCol;
					return true;
				}
			}
			// search the entire subsequent columns
			for (Row = A.GetUpperBound(0); Row >= 0; Row--)
			{
				for (Col = StartCol - 1; Col >= 0; Col--)
				{
					if (Math.Abs(A[Row, Col] - SearchVal) < MinDiff)
					{
						RowFound = Row;
						ColFound = Col;
						return true;
					}
				}
			}
			RowFound = -1;
			ColFound = -1;
			return false;
		}
		
		
		public static void MatSwapRows(ref double[,] A, int Row1, int Row2)
		{
			int Col;
			double X;
			
			if ((Row1 < 0) || (Row1 > (A.Length - 1) ) || (Row2 < 0) || (Row2 > (A.Length - 1)))
			{
				ReportError("Invalid row values", "MatSwapRows");
				return;
			}
			
			if (Row1 == Row2)
			{
				return;
			}
			
			try
			{
				
				for (Col = 0; Col <= A.GetUpperBound(1); Col++)
				{
					X = A[Row1, Col];
					A[Row1, Col] = A[Row2, Col];
					A[Row2, Col] = X;
				}
							
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatSwapRows");
			}
		}
		
		public static void MatSwapCols(ref double[,] A, int Col1, int Col2)
		{
			int Row;
			double X;
			
			if ((Col1 < 0) || (Col1 >  A.GetUpperBound(1)) || (Col2 < 0) || (Col2 > A.GetUpperBound(1)))
			{
				ReportError("Invalid row values", "MatSwapCols");
				return;
			}
			
			if (Col1 == Col2)
			{
				return;
			}
			
			try
			{
				
				for (Row = 0; Row <= A.GetUpperBound(0); Row++)
				{
					X = A[Row, Col1];
					A[Row, Col1] = A[Row, Col2];
					A[Row, Col2] = X;
				}				
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatSwapCols");
			}
		}
		
		public static double MatRowSum(ref double[,] A, int nRow)
		{
			double fSum;
			int nCol;
			
			fSum = mfBadDataVal;
			try
			{
				if ((nRow >= 0) && (nRow <= A.GetUpperBound(0)))
				{
					fSum = 0.0;
					for (nCol = 0; nCol <= A.GetUpperBound(1); nCol++)
					{
						fSum = fSum + A[nRow, nCol];
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatRowSum");
			}
            return fSum;
		}
		
		public static double MatColSum(ref double[,] A, int nCol)
		{
			double fSum;
			int nRow;
			
			fSum = mfBadDataVal;
			try
			{
				if ((nCol >= 0) && (nCol <= A.GetUpperBound(1)))
				{
					fSum = 0.0;
					for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
					{
						fSum = fSum + A[nRow, nCol];
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatColSum");
			}
            return fSum;
		}
		
		public static void MatColStats(ref double[,] A, int nCol, ref int n, ref double Mean, ref double Sdev)
		{
			double X;
			int nRow;
			double SumX = 0;
			double SumX2 = 0;
			
			Mean = mfBadDataVal;
			Sdev = mfBadDataVal;
			n = - 1;
			try
			{
				if ((nCol >= 0) && (nCol <= A.GetUpperBound(1)))
				{
					n = A.GetUpperBound(0) + 1;
					SumX = 0;
					SumX2 = 0;
					for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
					{
						X = A[nRow, nCol];
						SumX = SumX + X;
						SumX2 = SumX2 + X * X;
					}
				}
				
				Mean = SumX / n;
				Sdev = Math.Sqrt((SumX2 - SumX * SumX / n) /(n - 1));
				
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatColStats");
			}
		}
		
		public static double MatRowSumSquared(double[,] A, int nRow)
		{
			double fSum;
			int nCol;
			
			fSum = mfBadDataVal;
			try
			{
				if (nRow >= 0 && nRow <= A.GetUpperBound(0))
				{
					fSum = 0.0;
					for (nCol = 0; nCol <= A.GetUpperBound(1); nCol++)
					{
						fSum += Math.Pow(A[nRow, nCol], 2);
					}
				}				
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatRowSumSquared");
			}
			return fSum;
		}
		
		public static double MatColSumSquared(double[,] A, int nCol)
		{
			double fSum;
			int nRow;
			
			fSum = mfBadDataVal;
			try
			{
				if (nCol >= 0 && nCol <= A.GetUpperBound(1))
				{
					fSum = 0.0;
					for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
					{
						fSum += Math.Pow(A[nRow, nCol], 2);
					}
				}
			}
			catch (Exception ex)
			{
                ReportError(ex.Message, "MatColSumSquared");
			}
            return fSum;
        }
		
		public static double MatSumDiffTwoCols(double[,] A, int nCol1, int nCol2)
		{
			double fSum;
			int nRow;
			
			fSum = mfBadDataVal;
			try
			{
				if ((nCol1 >= 0) &&(nCol1 <= A.GetUpperBound(1)) && (nCol2 >= 0) &&(nCol2 <= A.GetUpperBound(1)))
				{
					fSum = 0.0;
					for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
					{
						fSum += A[nRow, nCol2] - A[nRow, nCol1];
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatSumDiffTwoCols");
			}
            return fSum;
		}
		
		public static double MatSumDiffTwoColsSquared(double[,] A, int nCol1, int nCol2)
		{
			double fSum;
			int nRow;
			
			fSum = mfBadDataVal;
			try
			{
				if ((nCol1 >= 0) &&(nCol1 <= A.GetUpperBound(1)) &&(nCol2 >= 0) &&(nCol2 <= A.GetUpperBound(1)))
				{
					fSum = 0.0;
					for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
					{
						fSum += Math.Pow(A[nRow, nCol2] - A[nRow, nCol1], 2);
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatSumDiffTwoColsSquared");
			}
            return fSum;
		}
				
		public static double MatRowMinMax(ref double[,] A, int nRow, ref double fMax)
		{
			int nCol;
			double fMin;
			
			fMax = mfBadDataVal;
			fMin = mfBadDataVal;
			try
			{
				
				if ((nRow >= 0) &&(nRow <= A.GetUpperBound(0)))
				{
					fMin = A[nRow, 0];
					fMax = fMin;
					for (nCol = 1; nCol <= A.GetUpperBound(1); nCol++)
					{
						
						if (fMin > A[nRow, nCol])
						{
							fMin = A[nRow, nCol];
						}
						
						if (fMax < A[nRow, nCol])
						{
							fMax = A[nRow, nCol];
						}
						
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatRowMinMax");
			}
			return fMin;
		}
		
		public static double MatColMinMax(ref double[,] A, int nCol, ref double fMax)
		{
			int nRow;
			double fMin;
			
			fMin = mfBadDataVal;
			fMax = mfBadDataVal;
			try
			{
				
				if ((nCol >= 0) &&(nCol <= A.GetUpperBound(1)))
				{
					fMin = A[0, nCol];
					fMax = fMin;
					for (nRow = 1; nRow <= A.GetUpperBound(0); nRow++)
					{
						
						if (fMin > A[nRow, nCol])
						{
							fMin = A[nRow, nCol];
						}
						
						if (fMax < A[nRow, nCol])
						{
							fMax = A[nRow, nCol];
						}
						
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatColMinMax");
			}
            return fMin;
		}
		
		
		public static double MatDiaglElemProduct(ref double[,] A)
		{
			double fProduct;
			int nRow;
			int nCol;
			int n;
			
			fProduct = 1;
			try
			{
                if (A.GetUpperBound(1) == A.GetUpperBound(0))
				{
					for (nRow = 0; nRow <= A.GetUpperBound(0); nRow++)
					{
						fProduct *= A[nRow, nRow];
					}
				}
				else
				{
					n = (Information.UBound(A, 2) > (A.Length - 1)) ? ((A.Length - 1) + 1) : (Information.UBound(A, 2) + 1);
					for (nRow = 0; nRow <= n - 1; nRow++)
					{
						for (nCol = 0; nCol <= n - 1; nCol++)
						{
							if (nRow == nCol)
							{
								fProduct *= A[nRow, nCol];
							}
						}
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatDiaglElemProduct");
			}
            return fProduct;
		}
		
		public static void VectToColumnMat(ref double[] AVect, ref double[,] A)
		{
			int i;

            A = new double[AVect.GetUpperBound(0) + 1, 1];
			for (i = 0; i < AVect.Length; i++)
			{
				A[i, 0] = AVect[i];
			}
		}
		
		public static void VectToRowMat(ref double[] AVect, ref double[,] A)
		{
			int i;

            A = new double[1, AVect.GetUpperBound(0) + 1];
			for (i = 0; i < AVect.Length; i++)
			{
				A[0, i] = AVect[i];
			}
		}
		
		public static double VectNorm(ref double[] AVect, int P)
		{
			int i;
			double fNorm;
			
			fNorm = 0;
			for (i = 0; i < AVect.Length; i++)
			{
				fNorm += Math.Pow(AVect[i], P);
			}
			fNorm = Math.Pow(fNorm, (1 / P));

            return fNorm;
		}
		
		
		public static double VectSum(ref double[] AVect)
		{
			int i;
			double Sum;
			
			Sum = 0;
			for (i = 0; i < AVect.Length; i++)
			{
				Sum += AVect[i];
			}
            return Sum;
		}
		
		public static double VectSumAbsVal(ref double[] AVect)
		{
			int i;
			double Sum;
			
			Sum = 0;
			for (i = 0; i < AVect.Length; i++)
			{
				Sum += Math.Abs(AVect[i]);
			}
            return Sum;
		}
		
		
		public static double VectProd(ref double[] AVect)
		{
			int i;
			double Prod;
			
			Prod = 1;
			for (i = 0; i < AVect.Length; i++)
			{
				Prod *= AVect[i];
			}
			return Prod;
		}
		
		public static double VectProdOfSum(ref double[] AVect, double X)
		{
			int i;
			double Prod;
			
			Prod = 1;
			for (i = 0; i < AVect.Length; i++)
			{
				Prod *= X + AVect[i];
			}
			return Prod;
		}
		
		public static double VectProdOfXminusVect(ref double[] AVect, double X)
		{
			int i;
			double Prod;
			
			Prod = 1;
			for (i = 0; i < AVect.Length; i++)
			{
				Prod *= X - AVect[i];
			}
			return Prod;
		}
		
		public static double VectProdOfVectMinusX(ref double[] AVect, double X)
		{
			int i;
			double Prod;
			
			Prod = 1;
			for (i = 0; i < AVect.Length; i++)
			{
				Prod *= AVect[i] - X;
			}
            return Prod;
		}
		
		public static double MatNorm(ref double[,] A, int P)
		{
			int i, j;
			double fNorm;
			
			fNorm = 0;
			for (i = 0; i <= A.GetUpperBound(0); i++)
			{
				for (j = 0; j <= A.GetUpperBound(1); j++)
				{
					fNorm += Math.Pow(A[i, j], P);
				}
			}
			fNorm = Math.Pow(fNorm, (1 / P));
            return fNorm;
		}
		
		public static void NormalizeVect(ref double[] AVect)
		{
			int i;
			double fNorm;
			
			fNorm = 0;
			for (i = 0; i < AVect.Length; i++)
			{
                fNorm += AVect[i] * AVect[i];
			}
			fNorm = Math.Sqrt(fNorm);
			for (i = 0; i < AVect.Length; i++)
			{
				AVect[i] = AVect[i] / fNorm;
			}
		}
		
		public static void MatAddMat(ref double[,] A, ref double[,] B, ref double[,] C)
		{
			int i, j;
			
			// C = new double[Information.UBound[A + 1, 1] + 1, Information.UBound[A + 1, 2] + 1];
            C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
			try
			{
				for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						C[i, j] = A[i, j] + B[i, j];
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatAddMat");
			}
		} //Add
		
		public static void MatAddMatInPlace(ref double[,] A, ref double[,] B)
		{
			int i, j;
			
			try
			{
                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						A[i, j] = A[i, j] + B[i, j];
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatAddMatInPlace");
			}
		} //Add
		
		
		public static void MatSubVal(double X, ref double[,] A, ref double[,] C)
		{
			int i, j;

            C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
			try
			{
                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						C[i, j] = A[i, j] - X;
					}
				}
			}
			catch (Exception ex)
			{
                C = new double[1, 1];
  			    C[0, 0] = mfBadDataVal;
				ReportError(ex.Message, "MatSubVal");
			}
		} //Sub
		
		public static void MatSubValInPlace(double X, ref double[,] A)
		{
            int i, j;
			
			try
			{
                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						A[i, j] = A[i, j] - X;
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatSubValInPlace");
			}
		} //Sub
		
		
		public static void MatSubMat(ref double[,] A, ref double[,] B, ref double[,] C)
		{
            int i, j;

            C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
			try
			{
                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <  A.GetUpperBound(1); j++)
					{
						C[i, j] = A[i, j] - B[i, j];
					}
				}
			}
			catch (Exception ex)
			{
			  C = new double[1, 1];
			  C[0, 0] = mfBadDataVal;
			  ReportError(ex.Message, "MatSubMat");				
			}
		}
		
		public static void MatSubMatInPlace(ref double[,] A, ref double[,] B)
		{
            int i, j;
			
			try
			{
				for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						A[i, j] -= B[i, j];
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatSubMatInPlace");
			}
		}
		
		
		public static void MatAddVal(ref double[,] A, double X, ref double[,] C)
		{
            int i, j;

            C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
			try
			{
                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						C[i, j] = A[i, j] + X;
					}
				}
			}
			catch (Exception ex)
			{
			  C = new double[1, 1];
			  C[0, 0] = mfBadDataVal;
			  ReportError(ex.Message, "MatAddVal");
			}
		}
		
		public static void MatAddValInPlace(ref double[,] A, double X)
		{
            int i, j;
			
			try
			{
				for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						A[i, j] += X;
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatAddValInPlace");
			}
		}
		
		public static void VectAddVect(ref double[] AVect, ref double[] BVect, ref double[] CVect)
		{
			int i;
			
			CVect = new double[AVect.GetUpperBound(0) + 1];
            for (i = 0; i <= AVect.GetUpperBound(0); i++)
			{
				CVect[i] = AVect[i] + BVect[i];
			}
		}
		
		public static void VectAddScaledVect(ref double[] AVect, double X, ref double[] BVect, ref double[] CVect)
		{
			int i;
			
			CVect = new double[AVect.GetUpperBound(0) + 1];
            for (i = 0; i <= AVect.GetUpperBound(0); i++)
			{
				CVect[i] = AVect[i] + X * BVect[i];
			}	
		}
		
		
		public static double VectTransMultVect(ref double[] AVect, ref double[] BVect)
		{
			double returnValue;
			double Res;
			int i;
			
			Res = 0;
            for (i = 0; i <= AVect.GetUpperBound(0); i++)
			{
				Res += AVect[i] * BVect[i];
			}
			returnValue = Res;
			return returnValue;
		}
		
		public static void VectMultVal(ref double[] AVect, double X, ref double[] BVect)
		{
			int i;

            BVect = new double[AVect.GetUpperBound(0) + 1];
            for (i = 0; i < AVect.GetUpperBound(0); i++)
			{
				BVect[i] = X * AVect[i];
			}
			
		}
		
		public static void VectMultValInPlace(ref double[] AVect, double X)
		{
			int i;

            for (i = 0; i <= AVect.GetUpperBound(0); i++)
			{
				AVect[i] *= X;
			}
			
		}
		
		public static void MatMultVal(ref double[,] A, ref double B, ref double[,] C)
		{
            int i, j;

            C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
			try
			{
                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						C[i, j] = B * A[i, j];
					}
				}
			}
			catch (Exception ex)
			{
			  C = new double[1, 1];
			  C[0, 0] = mfBadDataVal;
			  ReportError(ex.Message, "MatMultVal");				
			}
		}
		
		public static void MatMultValInPlace(ref double[,] A, double B)
		{
            int i, j;
			
			try
			{
				for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					for (j = 0; j <= A.GetUpperBound(1); j++)
					{
						A[i, j] *= B;
					}
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatMultValInPlace");
			}
		}
		
		
		public static void MatMultVect(ref double[,] A, ref double[] BVect, ref double[] CVect)
		{
			double X;
            int i, j;
			
			CVect = new double[BVect.Length];
			
			try
			{
				for (j = 0; j < BVect.Length; j++)
				{
					CVect[j] = mfBadDataVal;
				}
                if (A.GetUpperBound(1) != BVect.GetUpperBound(0))
				{
					ReportError("Incompatible matrix dimensions", "MatMultVect");
					return;
				}

                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					X = 0;
					for (j = 0; j < BVect.Length; j++)
					{
						X += BVect[j] * A[i, j];
					}
					CVect[i] = X;
				}
			}
			catch (Exception ex)
			{
			  CVect = new double[1];
			  CVect[0] = mfBadDataVal;
			  ReportError(ex.Message, "MatMultVect");				
			}
		}
		
		public static void MatMultVectInPlace(ref double[,] A, ref double[] BVect)
		{
			double X;
            int i, j;
            double[] C;
			
			C = new double[BVect.Length];
			try
			{
				for (j = 0; j < BVect.Length; j++)
				{
					C[j] = mfBadDataVal;
				}
                if (A.GetUpperBound(0) != BVect.GetUpperBound(0))
				{
					ReportError("Incompatible matrix dimensions", "MatMultVect");
					return;
				}

                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					X = 0;
					for (j = 0; j < BVect.Length; j++)
					{
						X += BVect[j] * A[i, j];
					}
					C[i] = X;
				}

				for (j = 0; j < BVect.Length; j++)
				{
					BVect[j] = C[j];
				}
			}
			catch (Exception ex)
			{
				ReportError(ex.Message, "MatMultVectInPlace");
			}
		}
		
		public static void MatMultMat(ref double[,] A, ref double[,] B, ref double[,] C)
		{
			double fSum;
            int i, j, k;

            C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
			try
			{
                for (i = 0; i <= A.GetUpperBound(0); i++)
				{
                    for (j = 0; j <= B.GetUpperBound(1); j++)
					{
						C[i, j] = mfBadDataVal;
					}
				}
                if (A.GetUpperBound(1) != B.GetUpperBound(0))
				{
					ReportError("Incompatible matrix dimensions", "MatMultMat)");
					return;
				}
				for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					for (j = 0; j <= B.GetUpperBound(1); j++)
					{
						fSum = 0;
						for (k = 0; k <= A.GetUpperBound(1); k++)
						{
							fSum += A[i, k] * B[k, j];
						}
						C[i, j] = fSum;
					}
				}
			}
			catch (Exception ex)
			{
              C = new double[1, 1]; 
              C[0 ,0] = mfBadDataVal;
              ReportError(ex.Message, "MatMultMat");				
			}
		}
		
		public static void MatMultMatInPlace(ref double[,] A, ref double[,] B)
		{
			double fSum;
            int i, j, k;
			double[,] C;
			
			C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
			try
			{
                if (A.GetUpperBound(0) != B.GetUpperBound(0))
				{
					ReportError("Incompatible matrix dimensions", "MatMultMat)");
					return;
				}
				for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					for (j = 0; j <= B.GetUpperBound(1); j++)
					{
						fSum = 0;
						for (k = 0; k <= A.GetUpperBound(1); k++)
						{
							fSum += A[i, k] * B[k, j];
						}
						C[i, j] = fSum;
					}
				}
				

				for (i = 0; i <= A.GetUpperBound(0); i++)
				{
					for (j = 0; j <= B.GetUpperBound(1); j++)
					{
						A[i, j] = C[i, j];
					}
				}
				return;
				
			}
			catch (Exception ex)
            {
                for (i = 0; i <= A.GetUpperBound(0); i++)
                {
                    for (j = 0; j <= B.GetUpperBound(1); j++)
                    {
                        C[i, j] = mfBadDataVal;
                    }
                }
                ReportError(ex.Message, "MatMultMatInPlace");
            }
        }


        public static void MatDivVal(ref double[,] A, double X, ref double[,] C)
        {
            int i, j;

            C = new double[A.GetUpperBound(0) + 1, A.GetUpperBound(1) + 1];
            try
            {
                for (i = 0; i <= A.GetUpperBound(0); i++)
                {
                    for (j = 0; j <= A.GetUpperBound(1); j++)
                    {
                        C[i, j] = A[i, j] / X;
                    }
                }
            }
            catch (Exception ex)
            {
              C = new double[1, 1];
              C[0, 0] = mfBadDataVal;
              ReportError(ex.Message, "MatDivVal");
            }
        }

        public static void MatDivValInPlace(ref double[,] A, double X)
        {
            int i;
            int j;

            try
            {
                for (i = 0; i <= A.GetUpperBound(0); i++)
                {
                    for (j = 0; j <= A.GetUpperBound(1); j++)
                    {
                        A[i, j] /= X;
                    }
                }
            }
            catch (Exception ex)
            {
                ReportError(ex.Message, "MatDivValInPlace");
            }
        }


        public static void MatTrans(ref double[,] A, ref double[,] C)
        {
            int i;
            int j;

            // redimension matrix C to have reverse dimensions of matrix A
            C = new double[A.GetUpperBound(1) + 1, A.GetUpperBound(0) + 1];
            try
            {
                // copy and transpose data into temp matrix X
                for (i = 0; i <= A.GetUpperBound(0); i++)
                {
                    for (j = 0; j <= A.GetUpperBound(1); j++)
                    {
                        C[j, i] = A[i, j];
                    }
                }
            }
            catch (Exception ex)
            {
              C = new double[1, 1];
              C[0, 0] = mfBadDataVal;
              ReportError(ex.Message, "MatTrans");
            }
        } //MatTrans

        public static void MatIdentity(ref double[,] A)
        {
            int i;
            int j;

            for (i = 0; i <= A.GetUpperBound(0); i++)
            {
                for (j = 0; j <= A.GetUpperBound(1); j++)
                {
                    A[i, j] = i == j ? 1 : 0;
                }
            }
        }

        public static void MatZero(ref double[,] A)
        {
            int i;
            int j;

            for (i = 0; i <= A.GetUpperBound(0); i++)
            {
                for (j = 0; j <= A.GetUpperBound(1); j++)
                {
                    A[i, j] = 0;
                }
            }
        }

        public static void MatRandom(ref double[,] A, double Xlow, double XHi)
        {
            int i;
            int j;
            Random r = new Random();

            //VBMath.Randomize(DateAndTime.Timer);
            for (i = 0; i <= A.GetUpperBound(0); i++)
            {
                for (j = 0; j <= A.GetUpperBound(1); j++)
                {
                    A[i, j] = Xlow + (XHi - Xlow) * r.NextDouble();
                }
            }
        }

        public static void MatRandomNormal(ref double[,] A, double Mean, double Sdev)
        {
            int i;
            int j;

            // VBMath.Randomize(DateAndTime.Timer);
            for (i = 0; i <= A.GetUpperBound(0); i++)
            {
                for (j = 0; j <= A.GetUpperBound(1); j++)
                {
                    A[i, j] = RandNorm(Mean, Sdev);
                }
            }
        }


        private static double RandNorm(double Mean, double Sdev)
        {
            double returnValue;
            // RandNorm returns a randomly distributed drawing from a
            // standard normal distribution i.e. one with:
            // Average = 0 and Standard Deviation = 1.0
            double fac;
            double rsq;
            double v1;
            double v2;

            // Each pass through the calculation of the routine produces
            //  two normally-distributed deviates, so we only need to do
            //  the calculations every other call. So we set the flag
            //  variable (to true) if gset contains a spare RandNorm value.
            if (RandNorm_flag)
            {
                returnValue = Mean + Sdev * RandNorm_gset;
                // Force calculation next time.
                RandNorm_flag = false;
            }
            else
            {
                // Don't have anything saved so need to find a pair of values
                // First generate a co-ordinate pair within the unit circle:
                do
                {
                    v1 = 2 * VBMath.Rnd(1) - 1.0;
                    v2 = 2 * VBMath.Rnd(1) - 1.0;
                    rsq = v1 * v1 + v2 * v2;
                } while (!(rsq <= 1.0));

                // Do the Math:
                fac = Math.Sqrt(-2.0 * Math.Log(rsq) / rsq);

                // Return one of the values and save the other (gset) for next time:
                return Mean + Sdev * (v2 * fac);
                //						RandNorm_gset = v1 * fac;
                //						RandNorm_flag = true;
            }

            return returnValue;
        }

        public static void MatFill(ref double[,] A, double X)
        {
            int i;
            int j;

            for (i = 0; i <= A.GetUpperBound(0); i++)
            {
                for (j = 0; j <= A.GetUpperBound(1); j++)
                {
                    A[i, j] = X;
                }
            }
        }

        public static void OpMult_Vect_VectT(ref double[] AVect, ref double[] BVect, ref double[,] CMat)
        {
            int i;
            int j;
            int n;

            n = AVect.Length - 1;
            if (n == (BVect.Length - 1))
            {
                CMat = new double[n + 1, n + 1];
                for (i = 0; i <= n; i++)
                {
                    for (j = 0; j <= n; j++)
                    {
                        CMat[i, j] = AVect[i] * BVect[j];
                    }
                }
            }
            else
            {
                CMat = new double[1, 1];
                CMat[0, 0] = -1.1E+30;
            }
        }

        public static double OpMult_VectT_Vect(ref double[] AVect, ref double[] BVect)
        {
            double Res;
            int i;
            int n;

            n = AVect.Length - 1;
            if (n == (BVect.Length - 1))
            {
                Res = 0;
                for (i = 0; i <= n; i++)
                {
                    Res += AVect[i] * BVect[i];
                }
            }
            else
            {
                Res = -1.1E+30;
            }
            return Res;
        }

        public static double OpMult_VectT_Mat_SameVect(ref double[] BVect, ref double[,] AMat)
        {
            double Res;
            double[] CVect;
            int i;
            int n;

            n = BVect.Length - 1;
            if (n == AMat.GetUpperBound(1) && n == AMat.GetUpperBound(1))
            {
                CVect = new double[n + 1];
                MatMultVect(ref AMat, ref BVect, ref CVect);
                Res = 0;
                for (i = 0; i <= n; i++)
                {
                    Res += CVect[i] * BVect[i];
                }
            }
            else
            {
                Res = -1.1E+30;
            }
            return Res;
        }

        public static void OpMult_Vect_VectT_Mat(ref double[] AVect, ref double[] BVect, ref double[,] AMat, ref double[,] CMat)
        {
            int n;
            double[,] BMat;

            BMat = new double[2, 2];
            n = AVect.Length - 1;
            if (n == BVect.GetUpperBound(0) && n == AMat.GetUpperBound(0) && n == AMat.GetUpperBound(1))
            {
                OpMult_Vect_VectT(ref AVect, ref BVect, ref BMat);
                MatMultMat(ref AMat, ref BMat, ref CMat);
            }
            else
            {
                CMat = new double[1, 1];
                CMat[0, 0] = -1.1E+30;
            }
        }

        public static void OpMult_Mat_Vect_VectT(ref double[] AVect, ref double[] BVect, ref double[,] AMat, ref double[,] CMat)
        {
            OpMult_Vect_VectT_Mat(ref AVect, ref BVect, ref AMat, ref CMat);
        }

        private static double Pythag(double A, double B)
				{
					double returnValue;
					double AbsA;
					double AbsB;
					
					AbsA = Math.Abs(A);
					AbsB = Math.Abs(B);
					if (AbsA > AbsB)
					{
						returnValue = AbsA * Math.Sqrt(1.0 + Math.Pow(AbsB / AbsA, 2));
					}
					else
					{
						returnValue = (AbsB == 0.0 ? 0.0 : (AbsB * Math.Sqrt(1.0 + Math.Pow(AbsA / AbsB, 2))));
					}
					return returnValue;
				}

        public static void LuDecomp(ref double[,] A, int nNumRows, ref int[] pivotIndex, ref double Factor)
				{
					const double Epsilon = 1.0e-20;
					double[] VV;
					int iMax, i, j, k;
					double big, dum, fSum, temp;
					
					iMax = 0;
					try
					{
						VV = new double[nNumRows];
						Factor = 1.0;
						for (i = 0; i < nNumRows; i++)
						{
							big = 0.0;
							for (j = 0; j < nNumRows; j++)
							{
								temp = Math.Abs(A[i, j]);
								if (temp > big) //ToDo: Unsupported feature: assignment within expression.
								{
									big = temp;
								}
							}
							if (big == 0.0)
							{
								ReportError("Singular matrix", "LuDecomp");
								return;
							}
							VV[i] = 1.0 / big;
						}
						for (j = 0; j < nNumRows; j++)
						{
							for (i = 0; i < j; i++)
							{
								fSum = A[i, j];
								for (k = 0; k < i; k++)
								{
									fSum -= A[i, k] * A[k, j];
								}
								A[i, j] = fSum;
							}
							big = 0.0;
							for (i = j; i < nNumRows; i++)
							{
								fSum = A[i, j];
								for (k = 0; k < j; k++)
								{
									fSum -= A[i, k] * A[k, j];
								}
								A[i, j] = fSum;
								dum = VV[i] * Math.Abs(fSum);
								if (dum >= big)
								{
									big = dum;
									iMax = i;
								}
							}
							if (j != iMax)
							{
								for (k = 0; k < nNumRows; k++)
								{
									dum = A[iMax, k];
									A[iMax, k] = A[j, k];
									A[j, k] = dum;
								}
								Factor = - Factor;
								VV[iMax] = VV[j];
							}
							pivotIndex[j] = iMax;
							if (A[j, j] == 0.0)
							{
								A[j, j] = Epsilon;
							}
							if (j != nNumRows - 1)
							{
								dum = 1.0 / A[j, j];
								for (i = j + 1; i < nNumRows; i++)
								{
									A[i, j] *= dum;
								}
							}
						}
					}
					catch (Exception ex)
					{
						ReportError(ex.Message, "LuDecomp");
					}
				}


        public static void LuBackSubst(ref double[,] A, int nNumRows, ref int[] pivotIndex, ref double[] BVect)
        {
            int i1, i ,j, ip;
            double fSum;

            i1 = 0;
            try
            {
                for (i = 0; i < nNumRows; i++)
                {
                    ip = pivotIndex[i];
                    fSum = BVect[ip];
                    BVect[ip] = BVect[i];
                    if (i1 != 0)
                    {
                        for (j = i1 - 1; j < i; j++)
                        {
                            fSum -= A[i, j] * BVect[j];
                        }
                    }
                    else if (fSum != 0.0)
                    {
                        i1 = i + 1;
                    }
                    BVect[i] = fSum;
                }
                for (i = nNumRows - 1; i >= 0; i--)
                {
                    fSum = BVect[i];
                    for (j = i + 1; j < nNumRows; j++)
                    {
                        fSum -= A[i, j] * BVect[j];
                    }
                    BVect[i] = fSum / A[i, i];
                }

                return;
            }
            catch (Exception ex)
            {
                ReportError(ex.Message, "LuBackSubst");
            }
        }

        public static void MV_LUBackSubst(ref double[,] A, ref int[] Index, int nNumRows, ref double[] B)
        {

            int i;
            int j;
            int k;
            int idx;
            double Sum;

            k = -1;
            for (i = 0; i < nNumRows; i++)
            {
                idx = Index[i];
                Sum = B[idx];
                B[idx] = B[i];
                if (k > -1)
                {
                    for (j = k; j < i; j++)
                    {
                        Sum -= A[i, j] * B[j];
                    }
                }
                if (Sum != 0)
                {
                    k = i;
                }
                B[i] = Sum;
            }

            for (i = nNumRows - 1; i >= 0; i--)
            {
                Sum = B[i];
                for (j = i + 1; j < nNumRows; j++)
                {
                    Sum -= A[i, j] * B[j];
                }
                B[i] = Sum / A[i, i];
            }
        }

        public static void MV_LUDecomp(ref double[,] A, ref int[] Index, int nNumRows)
        {

            int i;
            int j;
            int k;
            int iMax = 0;
            double large;
            double Sum;
            double z;
            double Z2;
            double[] scaleVect = new double[51];

            // loop to obtain the scaling element
            for (i = 0; i < nNumRows; i++)
            {
                large = 0;
                for (j = 0; j < nNumRows; j++)
                {
                    Z2 = Math.Abs(A[i, j]);
                    if (Z2 > large)
                    {
                        large = Z2;
                    }
                }
                // no non-zero large value? then exit with an error code
                if (large == 0)
                {
                    ReportError("Calculations error", "MV_LUDecomp");
                    return;
                }
                scaleVect[i] = 1 / large;
            }

            for (j = 0; j < nNumRows; j++)
            {
                for (i = 0; i <= j - 1; i++)
                {
                    Sum = A[i, j];
                    for (k = 0; k <= i - 1; k++)
                    {
                        Sum -= A[i, k] * A[k, j];
                    }
                    A[i, j] = Sum;
                }

                large = 0;

                for (i = j; i < nNumRows; i++)
                {
                    Sum = A[i, j];
                    for (k = 0; k <= j - 1; k++)
                    {
                        Sum -= A[i, k] * A[k, j];
                    }
                    A[i, j] = Sum;
                    z = scaleVect[i] * Math.Abs(Sum);
                    if (z >= large)
                    {
                        large = z;
                        iMax = i;
                    }
                }

                if (j != iMax)
                {
                    for (k = 0; k < nNumRows; k++)
                    {
                        z = A[iMax, k];
                        A[iMax, k] = A[j, k];
                        A[j, k] = z;
                    }
                    scaleVect[iMax] = scaleVect[j];
                }

                Index[j] = iMax;

                if (A[j, j] == 0)
                {
                    A[j, j] = 0.000000000000001;
                }

                if (j != nNumRows)
                {
                    z = 1 / A[j, j];
                    for (i = j + 1; i < nNumRows; i++)
                    {
                        A[i, j] *= z;
                    }
                }
            }
        }

        public static double MV_LUDeterminant(ref double[,] A, int nNumRows, int rowSwapFlag)
        {

            double result;
            int i;

            result = rowSwapFlag;
            for (i = 0; i < nNumRows; i++)
            {
                result = result * A[i, i];
            }

            return result;
        }

        public static void MV_LUInverse(ref double[,] A, ref double[,] InvA, ref int[] Index, int nNumRows)
        {

            double[] colVect;
            int i;
            int j;

            colVect = new double[nNumRows];
            for (j = 0; j < nNumRows; j++)
            {
                for (i = 0; i < nNumRows; i++)
                {
                    colVect[i] = 0;
                }
                colVect[j] = 1;
                MV_LUBackSubst(ref A, ref Index, nNumRows, ref colVect);
                for (i = 0; i < nNumRows; i++)
                {
                    InvA[i, j] = colVect[i];
                }
            }

        }

        public static double MatDeterminant(ref double[,] A, int nNumRows)
        {
            double returnValue;

            int[] Index;
            int i;
            int rowSwapFlag = 0;
            double result;
            int nErrCount;

            Index = new int[nNumRows + 1];
            nErrCount = GetErrorCount();
            MV_LUDecomp(ref A, ref Index, nNumRows);
            if (nErrCount != GetErrorCount())
            {
                returnValue = mfBadDataVal;
                return returnValue;
            }

            result = rowSwapFlag;

            for (i = 0; i < nNumRows; i++)
            {
                result *= A[i, i];
            }
            return result;
        }

        public static void MatInverse(ref double[,] A, int nNumRows)
        {

            double[] colVect;
            int[] Index;
            int i;
            int j;
            int nErrCount;

            colVect = new double[nNumRows];
            Index = new int[nNumRows];
            nErrCount = GetErrorCount();
            MV_LUDecomp(ref A, ref Index, nNumRows);
            if (nErrCount != GetErrorCount())
            {
                ReportError("Error", "MatInverse");
                return;
            }

            for (j = 0; j < nNumRows; j++)
            {
                for (i = 0; i < nNumRows; i++)
                {
                    colVect[i] = 0;
                }
                colVect[j] = 1;
                MV_LUBackSubst(ref A, ref Index, nNumRows, ref colVect);
                for (i = 0; i < nNumRows; i++)
                {
                    A[i, j] = colVect[i];
                }
            }

        }


        public static void SolveLU(ref double[,] A, ref double[] BVect, int nCount)
        {
            int[] pvtIdx;
            double Factor;
            int nInitErrCount;

            pvtIdx = new int[nCount];
            Factor = 1;
            nInitErrCount = GetErrorCount();

            LuDecomp(ref A, nCount, ref pvtIdx, ref Factor);
            // no new errors?
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuBackSubst(ref A, nCount, ref pvtIdx, ref BVect);
        }


        public static void LuImprove(ref double[,] A, ref double[,] ALud, ref int[] pivotIndex, ref double[] BVect, ref double[] Xvect, int nCount)
        {
            int i, j;
            double sdp;
            double[] R;

            R = new double[nCount];
            for (i = 0; i < nCount; i++)
            {
                sdp = -BVect[i];
                for (j = 0; j < nCount; j++)
                {
                    sdp += A[i, j] * Xvect[j];
                }
                R[i] = sdp;
            }
            LuBackSubst(ref ALud, nCount, ref pivotIndex, ref R);
            for (i = 0; i < nCount; i++)
            {
                Xvect[i] -= R[i];
            }
        }

        public static void SolveLU_Ex(ref double[,] A, ref double[] B, int nCount)
        {
            double[,] ACopy;
            double[] BCopy;
            int[] pvtIdx;
            double Factor;
            int nInitErrCount;

            Factor = 1;
            nInitErrCount = GetErrorCount();
            BCopy = new double[B.GetUpperBound(0) + 1];
            pvtIdx = new int[nCount];
            ACopy = new double[1, 1];
            ACopy[0, 0] = 0;

            DuplicateMat(ref A, ref ACopy);
            DuplicateVector(ref B, ref BCopy);

            LuDecomp(ref A, nCount, ref pvtIdx, ref Factor);
            // no new errors?
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuImprove(ref ACopy, ref A, ref pvtIdx, ref BCopy, ref B, nCount);
        } //SolveLU_Ex


        public static void SvdDecomp(ref double[,] A, int nNumRows, int nNumCols, ref double[] W, ref double[,] V)
        {
            const int MAX_ITERS = 30;
            double[] rv1;
            bool flag;
            int L;
            int nm;
            int IterCount;
            int i;
            int j;
            int m;
            int k;
            double C;
            double f;
            double h;
            double s;
            double X;
            double Y;
            double z;
            double g;
            double fScale;
            double Anorm;

            rv1 = new double[nNumRows];
            L = 0;
            nm = 0;
            g = 0;
            fScale = 0;
            Anorm = 0;

            for (i = 0; i < nNumRows; i++)
            {
                L = i + 2;
                rv1[i] = fScale * g;
                g = 0;
                s = 0;
                fScale = 0.0;
                if (i < nNumCols)
                {
                    for (k = i; k < nNumCols; k++)
                    {
                        fScale += Math.Abs(A[k, i]);
                    }
                    if (fScale != 0.0)
                    {
                        for (k = i; k < nNumCols; k++)
                        {
                            A[k, i] /= fScale;
                            s += Math.Pow(A[k, i], 2);
                        }
                        f = A[i, i];
                        g = -Sign(Math.Sqrt(s), f);
                        h = f * g - s;
                        A[i, i] = f - g;
                        for (j = L - 1; j < nNumRows; j++)
                        {
                            s = 0;
                            for (k = i; k < nNumCols; k++)
                            {
                                s += A[k, i] * A[k, j];
                            }
                            f = s / h;
                            for (k = i; k < nNumCols; k++)
                            {
                                A[k, j] += f * A[k, i];
                            }
                        }
                        for (k = i; k < nNumCols; k++)
                        {
                            A[k, i] *= fScale;
                        }
                    }
                }
                W[i] = fScale * g;
                g = 0;
                s = 0;
                fScale = 0.0;
                if (i + 1 <= nNumCols && i != nNumRows)
                {
                    for (k = L - 1; k < nNumRows; k++)
                    {
                        fScale = fScale + Math.Abs(A[i, k]);
                    }
                    if (fScale != 0.0)
                    {
                        for (k = L - 1; k < nNumRows; k++)
                        {
                            A[i, k] = A[i, k] / fScale;
                            s += A[i, k] * A[i, k];
                        }
                        f = A[i, L - 1];
                        g = -Sign(Math.Sqrt(s), f);
                        h = f * g - s;
                        A[i, L - 1] = f - g;
                        for (k = L - 1; k < nNumRows; k++)
                        {
                            rv1[k] = A[i, k] / h;
                        }
                        for (j = L - 1; j < nNumCols; j++)
                        {
                            s = 0;
                            for (k = L - 1; k < nNumRows; k++)
                            {
                                s += A[j, k] * A[i, k];
                            }
                            for (k = L - 1; k < nNumRows; k++)
                            {
                                A[j, k] += s * rv1[k];
                            }
                        }
                        for (k = L - 1; k < nNumRows; k++)
                        {
                            A[i, k] *= fScale;
                        }
                    }
                } //
                Anorm = Max(Anorm, Math.Abs(W[i]) + Math.Abs(rv1[i]));
            }
            for (i = nNumRows - 1; i >= 0; i--)
            {
                if (i < nNumRows - 1)
                {
                    if (g != 0.0)
                    {
                        for (j = L; j < nNumRows; j++)
                        {
                            V[j, i] = A[i, j] / A[i, L] / g;
                        }
                        for (j = L; j < nNumRows; j++)
                        {
                            s = 0;
                            for (k = L; k < nNumRows; k++)
                            {
                                s += A[i, k] * V[k, j];
                            }
                            for (k = L; k < nNumRows; k++)
                            {
                                V[k, j] += s * V[k, i];
                            }
                        }
                    }
                    for (j = L; j < nNumRows; j++)
                    {
                        V[i, j] = 0.0;
                        V[j, i] = 0.0;
                    }
                }
                V[i, i] = 1.0;
                g = rv1[i];
                L = i;
            }
            for (i = (nNumCols < nNumRows ? nNumCols : nNumRows) - 1; i >= 0; i--)
            {
                L = i + 1;
                g = W[i];
                for (j = L; j < nNumRows; j++)
                {
                    A[i, j] = 0.0;
                }
                if (g != 0.0)
                {
                    g = 1.0 / g;
                    for (j = L; j < nNumRows; j++)
                    {
                        s = 0;
                        for (k = L; k < nNumCols; k++)
                        {
                            s += A[k, i] * A[k, j];
                        }
                        f = s / A[i, i] * g;
                        for (k = i; k < nNumCols; k++)
                        {
                            A[k, j] += f * A[k, i];
                        }
                    }
                    for (j = i; j < nNumCols; j++)
                    {
                        A[j, i] *= g;
                    }
                }
                else
                {
                    for (j = i; j < nNumCols; j++)
                    {
                        A[j, i] = 0.0;
                    }
                }
                A[i, i] = A[i, i] + 1;
            }
            for (k = nNumRows - 1; k >= 0; k--)
            {
                for (IterCount = 0; IterCount <= 29; IterCount++)
                {
                    flag = true;
                    for (L = k; L >= 0; L--)
                    {
                        nm = L - 1;
                        if (Math.Abs(rv1[L]) + Anorm == Anorm)
                        {
                            flag = false;
                            break;
                        }
                        if (Math.Abs(W[nm]) + Anorm == Anorm)
                        {
                            break;
                        }
                    }
                    if (flag)
                    {
                        C = 0.0;
                        s = 1.0;
                        for (i = L - 1; i <= (k + 1) - 1; i++)
                        {
                            f = s * rv1[i];
                            rv1[i] = C * rv1[i];
                            if (Math.Abs(f) + Anorm == Anorm)
                            {
                                break;
                            }
                            g = W[i];
                            h = Pythag(f, g);
                            W[i] = h;
                            h = 1.0 / h;
                            C = g * h;
                            s = -f * h;
                            for (j = 0; j < nNumCols; j++)
                            {
                                Y = A[j, nm];
                                z = A[j, i];
                                A[j, nm] = Y * C + z * s;
                                A[j, i] = z * C - Y * s;
                            }
                        }
                    }
                    z = W[k];
                    if (L == k)
                    {
                        if (z < 0.0)
                        {
                            W[k] = -z;
                            for (j = 0; j < nNumRows; j++)
                            {
                                V[j, k] = -V[j, k];
                            }
                        }
                        break;
                    }
                    if (IterCount == MAX_ITERS)
                    {
                        ReportError("Failed to converge in " + IterCount + " IterCount", "SvdDecomp");
                        return;
                    }
                    X = W[L];
                    nm = k - 1;
                    Y = W[nm];
                    g = rv1[nm];
                    h = rv1[k];
                    f = ((Y - z) * (Y + z) + (g - h) * (g + h)) / (2.0 * h * Y);
                    g = Pythag(f, 1.0);
                    f = ((X - z) * (X + z) + h * (Y / (f + Sign(g, f)) - h)) / X;
                    C = 1.0;
                    s = 1.0;
                    for (j = L; j <= nm; j++)
                    {
                        i = j + 1;
                        g = rv1[i];
                        Y = W[i];
                        h = s * g;
                        g = C * g;
                        z = Pythag(f, h);
                        rv1[j] = z;
                        C = f / z;
                        s = h / z;
                        f = X * C + g * s;
                        g = g * C - X * s;
                        h = Y * s;
                        Y = Y * C;
                        for (m = 0; m < nNumRows; m++)
                        {
                            X = V[m, j];
                            z = V[m, i];
                            V[m, j] = X * C + z * s;
                            V[m, i] = z * C - X * s;
                        }
                        z = Pythag(f, h);
                        W[j] = z;
                        if (z != 0)
                        {
                            z = 1.0 / z;
                            C = f * z;
                            s = h * z;
                        }
                        f = C * g + s * Y;
                        X = C * Y - s * g;
                        for (m = 0; m < nNumCols; m++)
                        {
                            Y = A[m, j];
                            z = A[m, i];
                            A[m, j] = Y * C + z * s;
                            A[m, i] = z * C - Y * s;
                        }
                    }
                    rv1[L] = 0.0;
                    rv1[k] = f;
                    W[k] = X;
                }
            }
        }

        public static void SvdBackSubst(ref double[,] U, ref double[] W, ref double[,] V, ref double[] BVect, ref double[] Xvect, int nNumRows, int nNumCols)
        {
            int m;
            int j;
            int i;
            double s;
            double[] tmp;

            tmp = new double[nNumRows];
            Xvect = new double[nNumRows];
            for (j = 0; j < nNumRows; j++)
            {
                s = 0.0;
                if (W[j] != 0.0)
                {
                    for (i = 0; i < nNumCols; i++)
                    {
                        s += U[i, j] * BVect[i];
                    }
                    s /= W[j];
                }
                tmp[j] = s;
            }
            for (j = 0; j < nNumRows; j++)
            {
                s = 0.0;
                for (m = 0; m < nNumRows; m++)
                {
                    s += V[j, m] * tmp[m];
                }
                Xvect[j] = s;
            }
        }

        public static void SolveSvd(ref double[,] A, ref double[] BVect, ref double[] Xvect, int nCount)
        {
            double[,] V;
            double[] W;
            int nInitErrCount;

            W = new double[nCount];
            V = new double[nCount, nCount];
            nInitErrCount = GetErrorCount();

            SvdDecomp(ref A, nCount, nCount, ref W, ref V);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            SvdBackSubst(ref A, ref W, ref V, ref BVect, ref Xvect, nCount, nCount);
        } //SolveSvd

        public static void QrDecomp(ref double[,] A, ref double[] CVect, ref double[] Dvect, int nNumRows, ref bool bIsSingular)
        {
            int i;
            int j;
            int k;
            double fScale;
            double Sigma;
            double fSum;
            double Tau;

            bIsSingular = false;
            for (k = 0; k < nNumRows - 1; k++)
            {
                fScale = 0.0;
                for (i = k; i < nNumRows; i++)
                {
                    fScale = Max(fScale, Math.Abs(A[i, k]));
                }
                if (fScale == 0.0)
                {
                    bIsSingular = true;
                    CVect[k] = 0;
                    Dvect[k] = 0.0;
                }
                else
                {
                    for (i = k; i < nNumRows; i++)
                    {
                        A[i, k] = A[i, k] / fScale;
                    }
                    fSum = 0;
                    i = k;
                    while (i < nNumRows)
                    {
                        fSum += Math.Pow(A[i, k], 2);
                        i++;
                    }
                    Sigma = Sign(Math.Sqrt(fSum), A[k, k]);
                    A[k, k] = A[k, k] + Sigma;
                    CVect[k] = Sigma * A[k, k];
                    Dvect[k] = -fScale * Sigma;
                    for (j = k + 1; j < nNumRows; j++)
                    {
                        fSum = 0;
                        i = k;
                        while (i < nNumRows)
                        {
                            fSum += A[i, k] * A[i, j];
                            i++;
                        }
                        Tau = fSum / CVect[k];
                        for (i = k; i < nNumRows; i++)
                        {
                            A[i, j] -= Tau * A[i, k];
                        }
                    }
                }
            }
            Dvect[(nNumRows - 1)] = A[nNumRows - 1, nNumRows - 1];
            if (Dvect[(nNumRows - 1)] == 0.0)
            {
                bIsSingular = true;
            }
        }


        public static void QrSolve(ref double[,] A, ref double[] CVect, ref double[] Dvect, ref double[] BVect, int nNumRows)
        {
            int i;
            int j;
            double fSum;
            double Tau;

            for (j = 0; j < nNumRows - 1; j++)
            {
                fSum = 0;
                i = j;
                while (i < nNumRows)
                {
                    fSum += A[i, j] * BVect[i];
                    i++;
                }
                Tau = fSum / CVect[j];
                for (i = j; i < nNumRows; i++)
                {
                    BVect[i] -= Tau * A[i, j];
                }
            }
            RSolve(ref A, ref Dvect, ref BVect, nNumRows);
        }


        public static void RSolve(ref double[,] A, ref double[] d, ref double[] B, int nNumRows)
        {
            int i;
            int j;
            double fSum;

            B[nNumRows - 1] = B[nNumRows - 1] / d[nNumRows - 1];
            for (i = nNumRows - 2; i >= 0; i--)
            {
                fSum = 0;
                j = i + 1;
                while (j < nNumRows)
                {
                    fSum += A[i, j] * B[j];
                    j++;
                }
                B[i] = (B[i] - fSum) / d[i];
            }
        } //RSolve


        public static void Rotate(ref double[,] R, ref double[,] qt, int i, double A, double B, int nNumRows)
				{
					int j;
					double C;
					double fact;
					double s;
					double W;
					double Y;
					
					if (A == 0.0)
					{
						C = 0.0;
						s = B >= 0.0 ? 1.0 : - 1.0;
					}
					else if (Math.Abs(A) > Math.Abs(B))
					{
						fact = B / A;
						C = Sign(1.0 / Math.Sqrt(1.0 + Math.Pow(fact, 2)), A);
						s = fact * C;
					}
					else
					{
						fact = A / B;
						s = Sign(1.0 / Math.Sqrt(1.0 + Math.Pow(fact, 2)), B);
						C = fact * s;
					}
					for (j = i; j < nNumRows; j++)
					{
						Y = R[i, j];
						W = R[i + 1, j];
						R[i, j] = C * Y - s * W;
						R[i + 1, j] = s * Y + C * W;
					}
					for (j = 0; j < nNumRows; j++)
					{
						Y = qt[i, j];
						W = qt[i + 1, j];
						qt[i, j] = C * Y - s * W;
						qt[i + 1, j] = s * Y + C * W;
					}
				} //Rotate

        public static void QrUpdate(ref double[,] R, ref double[,] qt, ref double[] U, ref double[] V)
        {
            int i;
            int k;
            int n;

            n = (U.Length - 1) + 1;
            for (k = n - 1; k >= 0; k--)
            {
                if (U[k] != 0.0)
                {
                    break;
                }
            }
            if (k < 0)
            {
                k = 0;
            }
            for (i = k - 1; i >= 0; i--)
            {
                Rotate(ref R, ref qt, i, U[i], -U[(i + 1)], n);
                if (U[i] == 0.0)
                {
                    U[i] = Math.Abs(U[(i + 1)]);
                }
                else if (Math.Abs(U[i]) > Math.Abs(U[(i + 1)]))
                {
                    U[i] = Math.Abs(U[i]) * Math.Sqrt(1.0 + Square(U[(i + 1)] / U[i]));
                }
                else
                {
                    U[i] = Math.Abs(U[(i + 1)]) * Math.Sqrt(1.0 + Square(U[i] / U[(i + 1)]));
                }
            }
            for (i = 0; i <= n - 1; i++)
            {
                R[0, i] = R[0, i] + U[0] * V[i];
            }
            for (i = 0; i <= k - 1; i++)
            {
                Rotate(ref R, ref qt, i, R[i, i], -R[i + 1, i], n);
            }
        } //QrUpdate

        public static void SolveQr(ref double[,] A, ref double[] B, int nCount)
        {
            bool bIsSingular;
            int nInitErrCount;
            double[] C;
            double[] d;

            C = new double[nCount + 1];
            d = new double[nCount + 1];
            nInitErrCount = GetErrorCount();
            bIsSingular = false;

            QrDecomp(ref A, ref C, ref d, nCount, ref bIsSingular);
            if (GetErrorCount() > nInitErrCount || bIsSingular)
            {
                return;
            }
            QrSolve(ref A, ref C, ref d, ref B, nCount);
        } //SolveQr

        public static void GaussJordan(ref double[,] A, ref double[,] B, int nNumRows, int nNumCols)
        {
            int iCol;
            int iRow;
            int i;
            int j;
            int k;
            int L;
            int LL;
            double big;
            double dum;
            double pivinv;
            int[] pivotIndexc;
            int[] pivotIndexr;
            int[] ipiv;

            iCol = 0;
            iRow = 0;
            try
            {
                pivotIndexc = new int[nNumRows + 1];
                pivotIndexr = new int[nNumRows + 1];
                ipiv = new int[nNumRows + 1];

                for (j = 0; j < nNumRows; j++)
                {
                    ipiv[j] = 0;
                }
                for (i = 0; i < nNumRows; i++)
                {
                    big = 0.0;
                    for (j = 0; j < nNumRows; j++)
                    {
                        if (ipiv[j] != 1)
                        {
                            for (k = 0; k < nNumRows; k++)
                            {
                                if (ipiv[k] == 0)
                                {
                                    if (Math.Abs(A[j, k]) >= big)
                                    {
                                        big = Math.Abs(A[j, k]);
                                        iRow = j;
                                        iCol = k;
                                    }
                                }
                            }
                        }
                    }
                    ipiv[iCol] = ipiv[iCol] + 1;
                    if (iRow != iCol)
                    {
                        for (L = 0; L < nNumRows; L++)
                        {
                            Swap(ref A[iRow, L], ref A[iCol, L]);
                        }
                        for (L = 0; L < nNumCols; L++)
                        {
                            Swap(ref B[iRow, L], ref B[iCol, L]);
                        }
                    }
                    pivotIndexr[i] = iRow;
                    pivotIndexc[i] = iCol;
                    if (A[iCol, iCol] == 0.0)
                    {
                        ReportError("Singular Matrix", "GaussJordan");
                        return;
                    }
                    pivinv = 1.0 / A[iCol, iCol];
                    A[iCol, iCol] = 1.0;
                    for (L = 0; L < nNumRows; L++)
                    {
                        A[iCol, L] = A[iCol, L] * pivinv;
                    }
                    for (L = 0; L < nNumCols; L++)
                    {
                        B[iCol, L] *= pivinv;
                    }
                    for (LL = 0; LL < nNumRows; LL++)
                    {
                        if (LL != iCol)
                        {
                            dum = A[LL, iCol];
                            A[LL, iCol] = 0.0;
                            for (L = 0; L < nNumRows; L++)
                            {
                                A[LL, L] -= A[iCol, L] * dum;
                            }
                            for (L = 0; L < nNumCols; L++)
                            {
                                B[LL, L] -= B[iCol, L] * dum;
                            }
                        }
                    }
                }
                for (L = nNumRows - 1; L >= 0; L--)
                {
                    if (pivotIndexr[L] != pivotIndexc[L])
                    {
                        for (k = 0; k < nNumRows; k++)
                        {
                            Swap(ref A[k, pivotIndexr[L]], ref A[k, pivotIndexc[L]]);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
              ReportError(ex.Message, "GaussJordan");
            }
        }

        public static void SolveMultiRegLu(ref double[,] X, ref double[] Y, int nNumIndependentVars, ref double[] B)
        {
            double[,] Xt = new double[,] { { 1, 2 }, { 3, 4 } }; 
            double[,] StatSumMat = new double[,]{ { 1, 2 }, { 3, 4 } };
            int nCount;
            int[] pvtIdx;
            double Factor;
            int nInitErrCount;

            nCount = nNumIndependentVars + 1;
            pvtIdx = new int[nCount];
            B = new double[nCount];
            Factor = 1;
            nInitErrCount = GetErrorCount();
            MatTrans(ref X, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref X, ref StatSumMat);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref B);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuDecomp(ref StatSumMat, nCount, ref pvtIdx, ref Factor);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuBackSubst(ref StatSumMat, nCount, ref pvtIdx, ref B);
        } //SolveMultiRegLu


        public static void SolveMultiRegLu_Ex(ref double[,] X, ref double[] Y, int nNumIndependentVars, ref double[] B)
        {
            double[,] StatSumMatCopy = new double[,]{ { 1, 2 }, { 3, 4 } };
            double[,] Xt = new double[,]{ { 1, 2 }, { 3, 4 } };;
            double[,] StatSumMat = new double[,]{ { 1, 2 }, { 3, 4 } };;
            double[] BCopy;
            int nCount;
            int[] pvtIdx;
            double Factor;
            int nInitErrCount;

            nCount = nNumIndependentVars + 1;
            BCopy = new double[2];
            StatSumMat = new double[2, 2];
            StatSumMatCopy = new double[2, 2];
            pvtIdx = new int[nCount + 1];
            BCopy[0] = 0;
            Factor = 1;
            nInitErrCount = GetErrorCount();
            MatTrans(ref X, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref X, ref StatSumMat); // or (X, Xt, ...)
            DuplicateMat(ref StatSumMat, ref StatSumMatCopy);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref B); // was MatMultMat()
            DuplicateVector(ref B, ref BCopy);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuDecomp(ref StatSumMat, nCount, ref pvtIdx, ref Factor);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuImprove(ref StatSumMatCopy, ref StatSumMat, ref pvtIdx, ref BCopy, ref B, nCount);
        } //SolveMultiRegLu_Ex


        public static void SolvePolyRegLu(ref double[] X, ref double[] Y, int nOrder, ref double[] B)
        {
            double[,] Xt = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] XData = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] StatSumMat = new double[,] { { 1, 2 }, { 3, 4 } };
            int nCount;
            int[] pvtIdx;
            double Factor;
            int i;
            int j;
            int nInitErrCount;

            nCount = nOrder + 1;
            nInitErrCount = GetErrorCount();
            Factor = 1;
            XData = new double[X.GetUpperBound(0) + 1, nCount - 1 + 1];
            pvtIdx = new int[nCount - 1 + 1];

            for (i = 0; i <= (X.Length - 1); i++) // - 1
            {
                for (j = 0; j <= nCount - 1; j++)
                {
                    if (j == 0)
                    {
                        XData[i, j] = 1.0;
                    }
                    else if (j == 1)
                    {
                        XData[i, j] = X[i];
                    }
                    else
                    {
                        XData[i, j] = Math.Pow(X[i], j);
                    }
                }
            }

            MatTrans(ref XData, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref XData, ref StatSumMat);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref B);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuDecomp(ref StatSumMat, nCount, ref pvtIdx, ref Factor);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuBackSubst(ref StatSumMat, nCount, ref pvtIdx, ref B);
        } //SolvePolyRegLu


        public static void SolvePolyRegLu_Ex(ref double[] X, ref double[] Y, int nOrder, ref double[] B)
        {
            double[,] StatSumMatCopy = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] Xt = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] XData = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] StatSumMat = new double[,] { { 1, 2 }, { 3, 4 } };
            double[] BCopy;
            int nCount;
            int[] pvtIdx;
            double Factor;
            int i;
            int j;
            int nInitErrCount;


            nCount = nOrder + 1;
            BCopy = new double[2];
            BCopy[1] = 0;
            nInitErrCount = GetErrorCount();
            Factor = 1;
            XData = new double[X.GetUpperBound(0) + 1, nCount - 1 + 1];
            //ReDim BCopy(1)
            pvtIdx = new int[nCount - 1 + 1];

            for (i = 0; i < X.Length; i++)
            {
                for (j = 0; j < nCount; j++)
                {
                    if (j == 0)
                    {
                        XData[i, j] = 1.0;
                    }
                    else if (j == 1)
                    {
                        XData[i, j] = X[i];
                    }
                    else
                    {
                        XData[i, j] = Math.Pow(X[i], j);
                    }
                }
            }

            MatTrans(ref XData, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref XData, ref StatSumMat);
            DuplicateMat(ref StatSumMat, ref StatSumMatCopy);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref B);
            DuplicateVector(ref B, ref BCopy);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuDecomp(ref StatSumMat, nCount, ref pvtIdx, ref Factor);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            LuImprove(ref StatSumMatCopy, ref StatSumMat, ref pvtIdx, ref BCopy, ref B, nCount);
        }


        public static void SolveMultiRegSvd(ref double[,] X, ref double[] Y, int nNumIndependentVars, ref double[] B, ref double[] Coeff)
        {
            double[,] Xt = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] StatSumMat = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] V;
            double[] W;
            int nInitErrCount;
            int nCount;

            nCount = nNumIndependentVars + 1;
            W = new double[nCount - 1 + 1];
            V = new double[nCount + 1, nCount - 1 + 1];
            nInitErrCount = GetErrorCount();
            MatTrans(ref X, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref X, ref StatSumMat);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref B);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            SvdDecomp(ref StatSumMat, nCount, nCount, ref W, ref V);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            SvdBackSubst(ref StatSumMat, ref W, ref V, ref B, ref Coeff, nCount, nCount);
        }


        public static void SolvePolyRegSvd(ref double[] X, ref double[] Y, int nOrder, ref double[] B, ref double[] Coeff)
        {
            double[,] Xt = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] XData = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] StatSumMat = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] V;
            double[] W;
            int nInitErrCount;
            int nCount;
            int i;
            int j;

            nCount = nOrder + 1;
            W = new double[nCount - 1 + 1];
            V = new double[nCount + 1, nCount - 1 + 1];
            nInitErrCount = GetErrorCount();
            XData = new double[X.GetUpperBound(0) + 1, nCount - 1 + 1];

            for (i = 0; i < X.Length; i++)
            {
                for (j = 0; j < nCount; j++)
                {
                    if (j == 0)
                    {
                        XData[i, j] = 1.0;
                    }
                    else if (j == 1)
                    {
                        XData[i, j] = X[i];
                    }
                    else
                    {
                        XData[i, j] = Math.Pow(X[i], j);
                    }
                }
            }

            MatTrans(ref XData, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref XData, ref StatSumMat);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref B);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            SvdDecomp(ref StatSumMat, nCount, nCount, ref W, ref V);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            SvdBackSubst(ref StatSumMat, ref W, ref V, ref B, ref Coeff, nCount, nCount);
        }


        public static void SolveMultiRegQr(ref double[,] X, ref double[] Y, int nNumIndependentVars, ref double[] Coeff)
        {
            double[,] Xt = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] StatSumMat = new double[,] { { 1, 2 }, { 3, 4 } };
            double[] C;
            double[] d;
            bool bIsSingular;
            int nCount;
            int nInitErrCount;

            nCount = nNumIndependentVars + 1;
            nInitErrCount = GetErrorCount();
            bIsSingular = false;
            C = new double[nCount - 1 + 1];
            d = new double[nCount - 1 + 1];
            MatTrans(ref X, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref X, ref StatSumMat);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref Coeff);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            QrDecomp(ref StatSumMat, ref C, ref d, nCount, ref bIsSingular);
            if (GetErrorCount() > nInitErrCount || bIsSingular)
            {
                return;
            }
            QrSolve(ref StatSumMat, ref C, ref d, ref Coeff, nCount);
        } //SolveMultiRegQr


        public static void SolvePolyRegQr(ref double[] X, ref double[] Y, int nOrder, ref double[] Coeff)
        {
            double[,] Xt = new double[,] { { 1, 2 }, { 3, 4 } };
            double[,] XData;
            double[,] StatSumMat = new double[,] { { 1, 2 }, { 3, 4 } };
            double[] C;
            double[] d;
            bool bIsSingular;
            int nCount;
            int i;
            int j;
            int nInitErrCount;

            nCount = nOrder + 1;
            nInitErrCount = GetErrorCount();
            bIsSingular = false;
            C = new double[nCount - 1 + 1];
            d = new double[nCount - 1 + 1];
            XData = new double[X.GetUpperBound(0) + 1, nCount - 1 + 1];

            for (i = 0; i < X.Length; i++) // - 1
            {
                for (j = 0; j < nCount; j++)
                {
                    if (j == 0)
                    {
                        XData[i, j] = 1.0;
                    }
                    else if (j == 1)
                    {
                        XData[i, j] = X[i];
                    }
                    else
                    {
                        XData[i, j] = Math.Pow(X[i], j);
                    }
                }
            }

            MatTrans(ref XData, ref Xt);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultMat(ref Xt, ref XData, ref StatSumMat);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            MatMultVect(ref Xt, ref Y, ref Coeff);
            if (GetErrorCount() > nInitErrCount)
            {
                return;
            }
            QrDecomp(ref StatSumMat, ref C, ref d, nCount, ref bIsSingular);
            if (GetErrorCount() > nInitErrCount || bIsSingular)
            {
                return;
            }
            QrSolve(ref StatSumMat, ref C, ref d, ref Coeff, nCount);
        } //SolvePolyRegQr

        public static void Initialize()
        {
            ResetErrors(); // reset the error management data members
            mfBadDataVal = DEF_BAD_VAL; // set the default "BAD VALUE"
            // initialize data for bit shifting functions
            InitShiftData();
        }

    }
			
		}
